from . import node
from pseudonaja.c.PSymbolTable import Function, Variable
import pseudonaja.c.PInterpreter as pcint

class CallFunction(node.Node):

    def __init__(self, name, arglist, lineno):
        super().__init__(lineno)
        self.__name = name
        self.__arglist = arglist
      
    def interpret(self):
        stack_frame = {}

        # Internal functions don't need to be declared.
        if self.__name.lower() in ["div", "mod", "random", "len"]:
            
            arg1 = arg2 = None

            if self.__arglist and len(self.__arglist.args) > 0:
                arg1 = self.__arglist.args[0].value
                if len(self.__arglist.args) > 1:
                    arg2 = self.__arglist.args[1].value

            if self.__name.lower() == "div":
                return arg1 // arg2

            elif self.__name.lower() == "mod":
                return arg1 % arg2 

            elif self.__name.lower() == "random":
                from random import randint
                return randint(arg1, arg2) 

            else: # default array len command
                
                array_name = self.__arglist.args[0].name 
                try:
                    array_var = pcint.PInterpreter.symbols[array_name]
                except KeyError:
                    raise SyntaxError(f"Array name '{array_name}' undefined while calling len()")
                return array_var.end_index - array_var.start_index + 1
                
        else: # User defined functions

            func = pcint.PInterpreter.symbols[self.__name]
            if func.params:

                # get name of parameters from func in symbol table
                assert len(self.__arglist) == len(func.params), "length of args not equal to length of parameters"
                idx=0
                for arg in self.__arglist:
                    param = func.params[idx]
                    stack_frame[param.name] = Variable(param.name, param.type, arg.interpret())
                    idx +=1
                
                # Push stack frame
                pcint.PInterpreter.stack.append(stack_frame)

                # Run the procedure
                ret_value = func.statements.interpret()

                # Remove stack frame after procedure is complete
                pcint.PInterpreter.stack.pop() 
                return ret_value
                
            else: # Function with no arguments
                ret_value = func.statements.interpret()
                return ret_value

class Return(node.Node):

    def __init__(self, expr, lineno):
        super().__init__(lineno)
        self.__expr = expr

    def interpret(self):
        return self.__expr.interpret()

# Function declaration with arguments 
class FunctionDecl(node.Node):

  def __init__(self, name, params, type, statements, lineno):
    super().__init__(lineno)
    self.__name  = name
    self.__params  = params
    self.__type  = type
    self.__statements = statements

  def interpret(self):
    pcint.PInterpreter.symbols[self.__name] = Function(self.__name, self.__type, self.__params, self.__statements)


class Symbol:
    def __init__(self, name):
        self.__name = name

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, val):
        self.__name = val

class Variable(Symbol):
    def __init__(self, name, type, value=None):
        super().__init__(name)

        self.__value = value
        self.__type = type

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, value):
        self.__value = value

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return self.__str__()

class Array(Variable):
    def __init__(self, name, type, s_idx, e_idx):
        super().__init__(name, type, [None] * (e_idx - s_idx + 1))

        self.__s_idx = s_idx
        self.__e_idx = e_idx

    @property
    def start_index(self):
        return self.__s_idx

    @start_index.setter
    def start_index(self, value):
        self.__s_idx = value

    @property
    def end_index(self):
        return self.__e_idx

    @end_index.setter
    def end_index(self, value):
        self.__e_idx = value

    def __getitem__(self, idx):

        #print(f"Array __getitem_ called with index {idx}")

        if isinstance(idx, Variable):
            idx = idx.value

        if idx < self.__s_idx or idx > self.__e_idx:
            raise IndexError(f"Index out of range {self.name} [{idx}]")

        #print("Debug __getitem__ line 69 Symbol.py - value type is", type(self.value))

        return self.value[idx - self.__s_idx]

    def __setitem__(self, idx, value):

        #print(f"type of idx = {type(idx)}")
        #print(f"Array __setitem_ called with index {idx} and value {value}")

        if isinstance(idx, Variable):
            idx = idx.value

        if isinstance(value, Variable):
            value = value.value

        if idx < self.__s_idx or idx > self.__e_idx:
            raise IndexError(f"Index out of range {self.name} [{idx}]")

        if self.type == "INTEGER" and not isinstance(value, int):
            raise ValueError(f"Expecting {self.type} got {type(value)}")

        elif self.type == "STRING" and not isinstance(value, str):
            raise ValueError(f"Expecting {self.type} value got {type(value)}")

        self.value[idx - self.__s_idx] = value

    def __str__(self):
        return f"Array({self.name} {self.__s_idx}, {self.__e_idx}) {self.value}"

    def __repr__(self):
        return self.__str__()

class Function(Symbol):
    def __init__(self, name, params, type, statements):
        super().__init__(name)
        self.__params = params
        self.__type = type
        self.__statements = statements

    @property
    def params(self):
        return self.__params

    @property
    def type(self):
        return self.__type

    @property
    def statements(self):
        return self.__statements

    def __str__(self):
        return f"Function name={self.name}, params={self.params}, type={self.type}, statements={self.statements}"

    def __repr__(self):
        return self.__str__()
        
class Procedure(Symbol):
    def __init__(self, name, params, statements):
        super().__init__(name)

        self.__params = params
        self.__statements = statements

    @property
    def params(self):
        return self.__params

    @property
    def statements(self):
        return self.__statements

    def __str__(self):
        return f"Procedure name={self.name}, params={self.params}, statements={self.statements}"

    def __repr__(self):
        return self.__str__()

import pseudonaja.c.PInterpreter as pcint
class SymbolTable:
    def __init__(self):
        self.__table = {}
        self.keys = None
        self.kidx = -1

    def __setitem__(self, idx, val):
        self.__table [idx] = val

    def __getitem__(self, symbol):
        #print(f"call to __getitem__ with {symbol}")
        try:
            if symbol not in self.__table: # Global symbol table

                # Is there a call stack?
                if pcint.PInterpreter.stack and len(pcint.PInterpreter.stack) > 0 and isinstance(pcint.PInterpreter.stack[-1], dict):
                    return pcint.PInterpreter.stack[-1][symbol]
            else:
                return self.__table [symbol]
        except KeyError as e:
            raise SyntaxError(f"Symbol undefined {e}")

    def __iter__(self):
        self.keys = list(self.__table.keys())
        self.kidx=None
        if self.keys:
            self.kidx=0

        #print(f"__iter__ called {self.keys}")
        return self

    def __next__(self):
        #print(f"__next__called {self.kidx}")
        if self.keys and self.kidx < len(self.keys):
            key = self.keys[self.kidx]
            self.kidx += 1
            return key
        else:
            raise StopIteration

    def items(self):
        return self.__table.items()

    @property
    def table(self):
        return self.__table

if __name__ == "__main__":
    table = SymbolTable()
    a1 = Array("myarray1", "INTEGER", 10, 20)
    table[a1.name] = a1

    a2 = Array("myarray2", "INTEGER", 10, 20)
    table[a2.name] = a2

    if "myarray1" in table:
        print("myarray1 found")
    else:
        print("myarray1 not found")

    if "myarray" in table:
        print("myarray found")
    else:
        print("myarray not found")

    for i in range(10, 20):
        table[a.name][i] = str(i)

    for i in range(10, 20):
        print(table[a.name][i])


from sly import Parser
from pseudonaja.c.ast import identifier as identifier
from pseudonaja.c.ast import misc as misc
from pseudonaja.c.ast import literal as literal
from pseudonaja.c.ast import io as io
from pseudonaja.c.ast import binop as binop
from pseudonaja.c.ast import statement_list as statement_list
from pseudonaja.c.ast import selection as selection
from pseudonaja.c.ast import iteration as iteration
from pseudonaja.c.ast import procedure as procedure
from pseudonaja.c.ast import function as function

from pseudonaja.c import PLexer as PLexer


class PParser(Parser):

    root = None
    tokens = PLexer.PLexer.tokens

    # Un-comment the following line to output the parser logs for debugging any conflicts
    #debugfile = 'parser.out'

    # This sets the order of execution. The last values will have a higher precedence
    precedence = (
        ('right', ASSIGN  ),
        ('left',  OR      ),
        ('left',  AND     ),
        ('left',  EQUAL, NOTEQUAL),
        ('left',  BIGGER, SMALLER),
        ('left',  '+', '-'),
        ('left',  '*', '/'),
        ('right', UMINUS),
        ('left',  '.'     ), # We want to compute the float before do any operation

    )

    @_('statement_list')
    def program(self, p):
        self.root = p[0]

    @_('statement')
    def statement_list(self, p):
        return statement_list.StatementList(p[0], None)

    @_('statement statement_list')
    def statement_list(self, p):
        return statement_list.StatementList(p[0], p[1])

    @_('')
    def empty(self, p):
        pass

    @_('IDENTIFIER ":" TYPE')
    def identifier_type(self, p):
        return identifier.IdentifierDecl(p[0], p[2], p.lineno)

    @_('IDENTIFIER ":" ARRAY "[" NUMBER ":"  NUMBER "]" OF TYPE')
    def identifier_type(self, p):
        return identifier.ArrayIdentifierDecl(p[0], p[9], p[4], p[6], p.lineno)

    @_('identifier_type')
    def identifier_type_list(self, p):
        return [p[0]]

    @_('identifier_type "," identifier_type_list')
    def identifier_type_list(self, p):
        return [p[0]] + p[2]

    @_('empty')
    def identifier_type_list(self, p):
        pass

    # Definitions for statements start here -----------------------------------------------
    @_('DECLARE identifier_type')
    def statement(self, p):
        return identifier.Declare(p[1], p.lineno)

    # Declaring a procedure with 1 or more arguments
    @_('PROCEDURE IDENTIFIER "(" identifier_type_list ")" statement_list ENDPROCEDURE')
    def statement(self, p):
        return procedure.ProcedureDecl(p[1], p[3], p[5], p.lineno)

    # Declaring a procedure w/o arguments
    @_('PROCEDURE IDENTIFIER statement_list ENDPROCEDURE')
    def statement(self, p):
        return procedure.ProcedureDecl(p[1], None, p[2], p.lineno)

    # Declaring a function with 1 or more arguments
    @_('FUNCTION IDENTIFIER "(" identifier_type_list ")" RETURNS TYPE statement_list ENDFUNCTION')
    def statement(self, p):
        return function.FunctionDecl(p[1], p[6], p[3], p[7], p.lineno)

    # Declaring a function with zero arguments
    @_('FUNCTION IDENTIFIER RETURNS TYPE statement_list ENDFUNCTION')
    def statement(self, p):
        return function.FunctionDecl(p[1], p[3], None, p[4], p.lineno)

    # Declaring a procedure w/o arguments
    @_('CALL IDENTIFIER "(" arg_list ")"')
    def statement(self, p):
        return procedure.CallProcedure(p[1], p[3], p.lineno)

    # Declaring a procedure w/o arguments
    @_('CALL IDENTIFIER')
    def statement(self, p):
        return procedure.CallProcedure(p[1], None, p.lineno)

    @_('variable ASSIGN expression')
    def statement(self, p):
        return identifier.Assign(p[0], p[2], p.lineno)

    @_('INPUT IDENTIFIER')
    def statement(self, p):
        return io.Input(p[1],p.lineno)

    @_('OUTPUT arg_list')
    def statement(self, p):
        return io.Output(p[1], p.lineno)

    @_('IF expression THEN statement_list ENDIF')
    def statement(self, p):
        return selection.IfThen(p[1], p[3], p.lineno)

    @_('IF expression THEN statement_list ELSE statement_list ENDIF')
    def statement(self, p):
        return selection.IfThenElse(p[1], p[3], p[5], p.lineno)

    @_('CASE OF expression caselist ENDCASE')
    def statement(self, p):
        return selection.Case(p[2], p[3], p.lineno)

    @_('a_case')
    def caselist(self, p):
        return [p[0]]

    @_('a_case caselist')
    def caselist(self, p):
        return [p[0] ] + p[1]

    @_('constant ":" statement', 'OTHERWISE ":" statement')
    def a_case(self, p):
        return (p[0], p[2])

    @_('STRING', 'NUMBER')
    def constant(self, p):
        return p[0]

    @_('WHILE expression DO statement_list ENDWHILE')
    def statement(self, p):
        return iteration.While( p[1], p[3], p.lineno)

    @_('REPEAT statement_list UNTIL expression')
    def statement(self, p):
        return iteration.Repeat(p[1], p[3], p.lineno)

    @_('FOR IDENTIFIER ASSIGN expression TO expression statement_list NEXT')
    def statement(self, p):
        return iteration.For(p[1], p[3], p[5], None, p[6], p.lineno)

    @_('FOR IDENTIFIER ASSIGN expression TO expression STEP expression statement_list NEXT')
    def statement(self, p):
        return iteration.For(p[1], p[3], p[5], p[7], p[8], p.lineno)

    @_('RETURN expression')
    def statement(self, p):
        return function.Return(p[1], p.lineno)

    @_('"?" IDENTIFIER')
    def statement(self, p):
        #print(f"Got query for {p[1]}")
        return misc.QueryCommand(p[1], p.lineno)

    # Definitions for statements end here -------------------------------------------------

    # The operation is between a literal and another expression to allow chaining operators

    @_('"(" expression ")"')
    def expression(self, p):
        return p[1]
 
    @_('expresssion_operation')
    def expression(self, p):
        return p[0]

    @_('IDENTIFIER "(" arg_list ")"')
    def expression(self, p):
        return function.CallFunction(p[0], p[2], p.lineno)

    @_('IDENTIFIER "(" ")"')
    def expression(self, p):
        return function.CallFunction(p[0], None, p.lineno)

    @_('expression')
    def arg_list(self, p):
        return misc.ArgList(p[0], None)

    @_('expression "," arg_list')
    def arg_list(self, p):
        p[2] += misc.ArgList(p[0], None)
        return p[2]

    @_(
       'expression "-"      expression',
       'expression "+"      expression',
       'expression "/"      expression',
       'expression "*"      expression',
       'expression EQUAL    expression',
       'expression NOTEQUAL expression',
       'expression BIGGER   expression',
       'expression SMALLER  expression',
       'expression OR       expression',
       'expression AND      expression',
       )
    def expresssion_operation(self, p):
        #print(f"def expression_operation {p[0].name}, {p[0].type}")
        return binop.BinOp(p[0], p[1], p[2], p.lineno)


    @_('literal') # This must be added to avoid infinite recursion
    def expression(self, p):
        #print(f"found a lateral {p[0]}")
        return p[0]

    @_('"-" NUMBER %prec UMINUS')
    def literal(self, p):
        return literal.Literal(int(p[1]) * -1, "NUMBER", p.lineno)

    @_('NUMBER')  # Notice how we apply a number of grammar rules to the same method
    def literal(self, p):
        #print(f"Got literal INTEGER {p[0]}")
        return literal.Literal(int(p[0]), "NUMBER", p.lineno)

    @_('NUMBER "." NUMBER')
    def literal(self, p):
        return literal.Literal(float(f"{p[0]}.{p[2]}"), "NUMBER", p.lineno)

    @_('BOOL')
    def literal(self, p):
        return literal.Literal(True if p[0] == "TRUE" else False, "BOOLEAN", p.lineno)

    @_('STRING')  # Notice how we apply a number of grammar rules to the same method
    def literal(self, p):
        return literal.Literal(str(p[0]), "STRING", p.lineno)

    @_('IDENTIFIER "[" expression "]"')
    def literal(self, p):
        #print(f"Idenifier found on line {p.lineno}")
        return identifier.ArrayIdentifier(p[0], p[2], p.lineno)

    @_('IDENTIFIER')
    def literal(self, p):
        #print(f"Idenifier found on line {p.lineno}")
        return identifier.Identifier(p[0], p.lineno)

    @_('IDENTIFIER "[" expression "]"')
    def variable(self, p):
        return identifier.ArrayIdentifier(p[0], p[2], p.lineno)

    @_('IDENTIFIER')
    def variable(self, p):
        return identifier.Identifier(p[0], p.lineno)

    bin_op = {'+':'+', '-':'-', '*':'*', '/':'/', '>':'>', '>=':'>=', '<':'<',
              '<=':'<=', '=':'==', '<>':'!=', 'AND':'and', 'OR':'or'}

    def error(self, p):
        if p:
            raise SyntaxError(f"Parser error, line {p.lineno}, unexpected token type='{p.type}' value = '{p.value}'")
        else:
            raise SyntaxError("Unexpected end of file detected while parsing program")



import sys, os

sys.path += ['pseudonaja']
print(os.getcwd())

import pseudonaja.c.PInterpreter as pscint
import json

class RunTest:

    class Keyboard:
        def __init__(self, inputs):
            self.lines = inputs
            self.count = -1

        def readline(self):
            self.count += 1
            return self.lines[self.count]

    class Screen:
        def __init__(self):
            self.lines = []

        def write(self, line):
            #if line[-1] == '\n':
            #    line = line[:-1]
            self.lines.append(line)

        def flush(self):
            '''

            '''
            pass
        def __str__(self):
            return "".join(self.lines)

        def compare(self, name, outputs, screen):

            outputs = "".join(outputs)
            '''
            def dump(a, e):
                for i, c in enumerate(a):
                    print(f"{i} a={c} e={outputs[i] if i < len(outputs) else None}")

            dump(''.join(self.lines), outputs)
            '''
            assert ''.join(self.lines) == outputs, f"Test failed ({name})\n\nTest output\n{screen}\n\nExpected output\n{outputs}"
            print(f"Test ({name}) successful")


    def __init__(self, testfile):

        # read file
        with open(testfile+".json", 'r') as t:
            data = t.read()

        # parse file
        test = json.loads(data)

        import sys
        sys.stdin = RunTest.Keyboard(test['inputs'])

        stdout_save = sys.stdout
        s = RunTest.Screen()
        sys.stdout = s

        pscint.PInterpreter().run('\n'.join(test['code']))

        sys.stdout = stdout_save

        s.compare(test['name'], test['outputs'], s)


#print("DEBUG", os.getcwd())
if __name__ == "__main__":
    RunTest("pseudonaja/tests/Test4")


from . import node
from . import misc
import pseudonaja.c.PInterpreter as pcint
from pseudonaja.c.PSymbolTable import Variable, Array


class Assign(node.Node):
    '''
    Assign an expression to a variable
    '''
    def __init__(self, var, expr, lineno):
        super().__init__(lineno)

        self.__var = var
        self.__expr = expr

    def interpret(self):

        #print("Debug: Assign.interpret()", self.__var)

        if self.__var.name not in pcint.PInterpreter.symbols:
            #print(pcint.PInterpreter.symbols.table)
            raise SyntaxError(f"Symbol '{self.__var.name}' undefined on line {self.lineno}")

        if isinstance(self.__var, ArrayIdentifier):

            name = self.__var.name
            idx  = self.__var.idx

            assert name and idx, f"Assertion failed: Array assignment error: {name}[{idx}]" 

            #print(f"Got array idenifier: name={name} index={idx}")
            pcint.PInterpreter.symbols[name][idx] = self.__expr.interpret()

        elif   isinstance(self.__var, Identifier):
            value = self.__expr.interpret()
            var_type = pcint.PInterpreter.symbols[self.__var.name].type
 
            pcint.PInterpreter.symbols[self.__var.name].value = misc.type_cast(var_type, value)
        else:
            raise SyntaxError(f"Assign.interpret: ID type undefined {self.__var}")


class Declare(node.Node):

    def __init__(self, decl, lineno):

            super().__init__(lineno)
            self.__declaration = decl

    def interpret(self):
        self.__declaration.interpret()

    def __str__(self):
        return f"Declare: '{self.__declaration}'"

    def __repr__(self):
        return self.__str__()


class Identifier(node.Node):

    def __init__(self, name, lineno):
        super().__init__(lineno)
        self.__name = name

    @property
    def name(self):
        return self.__name

    @property
    def type(self):
        pcint.PInterpreter.symbols[ self.name ].type

    @property
    def value(self):
        return self.interpret()

    def interpret(self):
 
        # check if identifier in the global symbol table
        if self.name not in pcint.PInterpreter.symbols:

            # check if there is a stack frame, if so check if name is in the stack frame
            if len(pcint.PInterpreter.stack) == 0 or isinstance(pcint.PInterpreter.stack[-1], dict) and self.name not in pcint.PInterpreter.stack[-1]: 

                raise SyntaxError(f"Identifier '{self.name}' is undefined")

            else: # found identifier in stack frame
                value = pcint.PInterpreter.stack[-1][self.name]

        else:
            value = pcint.PInterpreter.symbols[ self.name ].value

        return value


class ArrayIdentifier(Identifier):

    def __init__(self, name, expr, lineno):
        super().__init__(name, lineno)
        self.__expr = expr

    @property
    def name(self):
        return super().name

    @property
    def idx(self):
        #print(f"DEBUG: ArrayIdentifier - idx property {type(self.__expr)}")
        return self.__expr.interpret()

    def interpret(self):
        return pcint.PInterpreter.symbols[ self.name ] [self.idx]

    def __str__(self):
        return f"ArrayIdentifier {super().name}[ {self.__expr.interpret()} ]"

    def __repr__(self):
        return self.__str__()


class IdentifierDecl(node.Node):

    def __init__(self, name, id_type, lineno):
        super().__init__(lineno)
        self.__name = name
        self.__type = id_type

    @property
    def name(self):
        return self.__name
    @property
    def type(self):
        return self.__type

    def interpret(self):
        pcint.PInterpreter.symbols[self.__name] = Variable(self.__name, self.__type)

class ArrayIdentifierDecl(IdentifierDecl):

    def __init__(self, name, type, start, end, lineno):
        super().__init__(name, type, lineno)
        self.__start_idx = start
        self.__end_idx = end

    def interpret(self):

        __name, __type  = self.name, self.type
        pcint.PInterpreter.symbols[__name] = Array(__name, __type, self.__start_idx, self.__end_idx)



import sys
import pseudonaja.c.PInterpreter as pcint

if len(sys.argv) == 1:
    pcint.PInterpreter().repl()
else:
    for file in sys.argv[1:]:
        try:
            with open(file, "r") as prog:
                prog = prog.read()
                pcint.PInterpreter().run(prog)
        except FileNotFoundError as e:
            print(f"{e}")


from . import node

class IfThen(node.Node):

  def __init__(self, condition, statements, lineno):
    super().__init__(lineno)
    self.__condition  = condition
    self.__statements = statements

  def interpret(self):
    if self.__condition.interpret():
        self.__statements.interpret()

class IfThenElse(node.Node):

  def __init__(self, condition, statements1, statements2, lineno):
    super().__init__(lineno)
    self.__condition  = condition
    self.__statements1 = statements1
    self.__statements2 = statements2

  def interpret(self):
    if self.__condition.interpret():
        self.__statements1.interpret()
    else:
        self.__statements2.interpret()

class Case(node.Node):

    def __init__(self, expr, case_list, lineno):
        super().__init__(lineno)
        self.__expr = expr
        self.__case_list = case_list

    def interpret(self):
        expr_val = self.__expr.interpret()
        found, otherwise = False, -1
        for i, c in enumerate(self.__case_list):
            literal, statements = c
            if literal == 'OTHERWISE':
                otherwise = i
            if expr_val == literal:
                statements.interpret()
                found = True
                break
        if not found and otherwise != -1:
            self.__case_list[otherwise][1].interpret()




from . import node
from pseudonaja.c.PSymbolTable import Variable, Procedure
from pseudonaja.c import PInterpreter as pcint

class CallProcedure(node.Node):

    def __init__(self, name, arglist, lineno):
        super().__init__(lineno)
        self.__name = name
        self.__arglist = arglist
      
    def interpret(self):
        proc = pcint.PInterpreter.symbols[self.__name]
        if proc.params:
            stack_frame = {}
            # get name of parameters from proc in symbol table
            assert len(self.__arglist) == len(proc.params), "length of args not equal to length of parameters"
            idx=0
            for arg in self.__arglist:
                param = proc.params[idx]
                stack_frame[param.name] = Variable(param.name, param.type, arg.interpret())
                idx +=1
            
            # Push stack frame
            pcint.PInterpreter.stack.append(stack_frame)

            # Run the procedure
            proc.statements.interpret()

            # Remove stack frame after procedure is complete
            pcint.PInterpreter.stack.pop() 
            
        else:
            proc.statements.interpret()

class ProcedureDecl(node.Node):

  def __init__(self, name, args, statements, lineno):
    super().__init__(lineno)
    self.__name  = name
    self.__args = args
    self.__statements = statements

  def interpret(self):
    pcint.PInterpreter.symbols[self.__name] = Procedure(self.__name, self.__args, self.__statements)



from abc import ABC, abstractmethod

class Node(ABC):
    '''
    This is the base class for all nodes of the Abstract Syntax Tree.
    This ensure that all subclasses implement the interpret() nmethod.
    It also includes an attribute, lineno, that is used to track where
    parts of the pseudocode program lines are parsed. 
    '''

    def __init__(self, lineno=-1):
        self.lineno = lineno

    @abstractmethod
    def interpret(self):
        pass

    @property
    def lineno(self):
        return self.__lineno

    @lineno.setter
    def lineno(self, l):
        self.__lineno = l



from . import node
from pseudonaja.c import PInterpreter as pcint
from pseudonaja.c.PSymbolTable import Variable

class Repeat(node.Node):

  def __init__(self, statements, condition, lineno):
    super().__init__(lineno)
    self.__statements = statements
    self.__condition  = condition

  def interpret(self):
    while True:
        self.__statements.interpret()
        if self.__condition.interpret():
            break

class While(node.Node):

  def __init__(self, condition, statements, lineno):
    super().__init__(lineno)
    self.__condition  = condition
    self.__statements = statements

  def interpret(self):
    while self.__condition.interpret():
        self.__statements.interpret()

class For(node.Node):

  def __init__(self, identifier, start, end, step, statements, lineno):
    super().__init__(lineno)
    self.__identifier  = identifier
    self.__start = start
    self.__end = end
    self.__step = step
    self.__statements = statements

  def interpret(self):

    stack_frame = {}

    stack_frame[self.__identifier] = Variable(self.__identifier, "INTEGER")

    # Push stack frame
    pcint.PInterpreter.stack.append(stack_frame)

    step = 1 if self.__step == None else self.__step.interpret()
    for i in range(self.__start.interpret(), self.__end.interpret()+1, step):

        stack_frame[self.__identifier].value = i
        
        #print(f"DEBUG: For.interpret() {stack_frame[self.__identifier].value} = {i} ")
        # Run the procedure
        self.__statements.interpret()

    # Remove stack frame after procedure is complete
    pcint.PInterpreter.stack.pop() 




from . import node
from . import function

class StatementList(node.Node):
    '''
    A pseudocode program consists of a list of statements
    This is either a single statement, or a statement followed by another statement
    A statement list is built by appending statements to self.__statements
    '''
    def __init__(self, s1, s2):
        self.__statements = [ s1 ]
        if s2:
            self.__statements += s2.statements

    @property
    def statements(self):
        return self.__statements

    def interpret(self):
        for statement in self.statements:
            if isinstance(statement, function.Return):
                return statement.interpret()
            else:
                statement.interpret()



from sly import Lexer

class PLexer(Lexer):
    
    # This is the set of tokens we are exporting to the Parser
    tokens =    {
                    IDENTIFIER, STRING, NUMBER, OR, AND, REPEAT, UNTIL,
                    IF, THEN, ELSE, ENDIF, WHILE, DO, ENDWHILE, ASSIGN, OUTPUT, INPUT,
                    DECLARE, TYPE, BOOL, BIGGER, SMALLER, EQUAL, NOTEQUAL, PROCEDURE, ENDPROCEDURE,
                    ARRAY, OF, CALL, FUNCTION, ENDFUNCTION, RETURNS, RETURN, FOR, TO, STEP, NEXT, CASE, ENDCASE,
                    OTHERWISE,
                }

    # Any literals we did not define as tokens, will be available for usage in the Parser
    literals = {'+', '-', '/', '*', '.', '!', '(', ')', ',', ':', '?', '[', ']'}

    # Any literals we want to ignore
    ignore = ' \t'

    # The definition of each token in a regex pattern - Notice that the order MATTERS!! First match will be taken
    OR       = r'OR'
    AND      = r'AND'
    ASSIGN   = r'(<-)'
    BOOL     = r'TRUE|FALSE'
    NOTEQUAL = r'<>'
    BIGGER   = r'>=|>'
    SMALLER  = r'<=|<'
    EQUAL    = r'='

    TYPE     = r'(INTEGER|REAL|CHAR|STRING|BOOLEAN|DATE)'

    # This decorator allows us to add a logic before returning the matched token.
    @_(r"(0|[1-9][0-9]*)")
    def NUMBER(self, t):
        t.value = int(t.value)
        return t

    @_(r'''("[^"\\]*(\\.[^"\\]*)*"|'[^'\\]*(\\.[^'\\]*)*')''')
    def STRING(self, t):
        t.value = self.remove_quotes(t.value)
        return t

    # Notice Identifier comes after string because most words in a string would be matched with the identifier pattern
    IDENTIFIER = r'[a-zA-Z_][a-zA-Z0-9_]*'

    # define keyword tokens and make them case insensitive
    keywords = ['OUTPUT', 'IF', 'THEN', 'ELSE', 'ENDIF', 'WHILE', 'DO', 'ENDWHILE',
                'REPEAT', 'UNTIL', 'INPUT', 'DECLARE', 'PROCEDURE', 'ENDPROCEDURE',
                'ARRAY', 'OF', 'CALL', 'FUNCTION', 'ENDFUNCTION', 'RETURNS',
                'RETURN', 'FOR', 'TO', 'STEP', 'NEXT', 'CASE', 'ENDCASE',
                'OTHERWISE']

    def IDENTIFIER(self, t):
        if t.value.upper() in self.keywords :
            t.type = t.value = t.value.upper()

        elif t.value.upper() in ['INTEGER', 'REAL', 'CHAR', 'STRING', 'BOOLEAN', 'DATE']:
            t.type = 'TYPE'
            t.value = t.value.upper()

        elif t.value.upper() in ['TRUE', 'FALSE']:
            t.type = 'BOOL'
            t.value = t.value.upper()

        elif t.value.upper() == 'AND':
            t.type = t.type = 'AND'

        elif t.value.upper() == 'OR':
            t.type = t.value = 'OR'

        elif t.value.upper() == 'OTHERWISE':
            t.type = t.value = 'OTHERWISE'

        return t

    @_(r'//.*')
    def ignore_comment(self, t):
        pass

    @_(r'\n+')
    def ignore_newline(self, t):
        self.lineno += t.value.count('\n')

    def remove_quotes(self, text: str):
        if text.startswith('\"') or text.startswith('\''):
            return text[1:-1]
        return text

    def error(self, t):
        print("Illegal character '%s'" % t.value[0])
        self.index += 1


from . import node  
import pseudonaja.c.PInterpreter as pcint

class QueryCommand(node.Node):
  def __init__(self, cmd, lineno):
    super().__init__(lineno)
    self.__command = cmd

  def interpret(self):
    if self.__command == "symboltable":
        print("Dumping symbol table");
        for n, v in pcint.PInterpreter.symbols.items():
            print(f"name='{n}', value='{v}'")

def type_cast(var_type, value):

    val = None
    #Cast the input to the type of the variable
    try:
        if var_type == "INTEGER":
            val = int(value)

        elif var_type == "REAL":
            val = float(value)

        elif var_type == "CHAR":
            val = value[0]

        elif var_type == "BOOLEAN":
            val = value

        elif var_type == "STRING":
            val = str(value)

        elif var_type == "DATE":
            import datetime
            if "/" in value:
                val = value.split("/")
            elif "-" in value:
                val = value.split("-")
            else:
                raise ValueError(f"Invalid date format {value} use dd/mm/yyyy")

            try:
                val = datetime.datetime(int(val[2]), int(val[1]), int(val[0]))
            except ValueError as e:
                raise ValueError(f"Invalid date {val} - {e}")
        else:
            raise ValueError(f"Invalid data type {val}")

    except ValueError as e:
        raise SyntaxError(e)
    
    return val

class ArgList(node.Node):

    def __init__(self, arg1, arg2=None, lineno=-1):
        super().__init__(lineno)
        self.__args = [arg1]
        if arg2 != None:
            self.__args += [arg2]

    def __iadd__(self, arg):
        #print("Debug: __iadd__ called")
        self.__args = arg.args + self.__args
        return self

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.__args):
            val = self.__args[self.n]
            self.n += 1
            return val
        else:
            raise StopIteration

    def __len__(self):
        return len(self.__args)

    @property
    def args(self):
        return self.__args

    def interpret(self):
        return self.args




from . import node

class Literal(node.Node):

    def __init__(self, val, type, lineno):

        super().__init__(lineno)
        self.__value = val
        self.__type = type

    @property
    def value(self):
        return self.__value

    @property
    def type(self):
        return self.__type

    def interpret(self):
        return self.__value

    def __repr__(self):
        return str(self.__value)




from pseudonaja.c.PLexer  import PLexer
from pseudonaja.c.PParser import PParser
from pseudonaja.c.PSymbolTable import SymbolTable

class UnableToContinue(Exception):
    def __init__(self, *args):
        self.__message = args[[0]]
    
    def __str__(self):
        return self.__message

class PInterpreter:

    symbols = SymbolTable()
    stack   = []

    def __init__(self):
        self.lexer = PLexer()
        self.parser = PParser()


    def repl(self):
        prog = []
        finished = False
        while not finished:
            line = input ("Pseudonaja.c>> ")
            if line:
                if line[0] == '.':
                    line = line.strip()
                    if len(line) == 1:
                        if len(prog) > 0:
                            try:
                                prog = "\n".join(prog)
                                self.run(prog)
                                prog = []
                            except Exception as e:
                                print(f"Error while running code {e}")
                        else:
                            print("There is nothing to run")
                    elif line == '.quit':
                        finished = True
                    elif line == '.list':
                        for l in prog:
                            print(l)
                    else:
                        print(f"Unrecognised '.' command {line}")
                else:
                    prog.append(line)

    def run(self, program):
        try:
            tokens = self.lexer.tokenize(program)
            self.parser.parse(tokens)
            self.parser.root.interpret()

        except SyntaxError as e:
            print(f"\nSyntax Error: {e}")

        '''
        except UnableToContinue as e:
            print(f"{e}")

        except TypeError as e:
            print(f"{e}")
            
        except ValueError as e:
            print(f"{e}")

        except AssertionError as e:
            print(f"{e}")
        '''


from . import node
from . import misc
import pseudonaja.c.PInterpreter as pcint

class Input(node.Node):

    def __init__(self, identifier, lineno):

        #print(f"Debug: Output.__init__ {args}")

        super().__init__(lineno)
        self.__identifier = identifier

    def interpret(self):

        # Check to see if the vriable has been declared
        if self.__identifier in pcint.PInterpreter.symbols:

            # Get the variable from the symbol table
            var = pcint.PInterpreter.symbols[self.__identifier]
            val = input()
            while val == "":
                val = input()

            val = misc.type_cast(var.type, val)

            # Set variable's value to the value returned from input
            var.value = val
        else:
            print(f"Error: symbol '{self.__identifier}' undefined on line {self.lineno}")

class Output(node.Node):

    def __init__(self, args, lineno):

        #print(f"Debug: Output.__init__ {args}")

        super().__init__(lineno)

        self.__args = args

    def interpret(self):

        #print(f"Debug: Output.interpret {self.__args}")

        for arg in self.__args:
            print(arg.interpret(), end=" ")

        print()



from . import node
from . import identifier
import pseudonaja.c.PParser as PParser
from pseudonaja.c.PSymbolTable import Variable

class BinOp(node.Node):

    def __init__(self, left, op, right, lineno):

        #print("BinOp().__init__() : Got here")
        super().__init__(lineno)
        self.__left = left
        self.__op = PParser.PParser.bin_op[op.upper()]
        self.__right = right

    def interpret(self):
        from datetime import datetime, timedelta

        if self.__op.lower() in ['and', 'or'] and (isinstance(self.__left, BinOp) ^ isinstance(self.__right, BinOp)):
            raise TypeError(f"{self.__left} '{self.__op}' '{self.__right}' on line {self.lineno}")

        if not isinstance(self.__left, identifier.Identifier) and isinstance(self.__right, identifier.Identifier) and not isinstance(self.__left, BinOp) and not isinstance(self.__right, BinOp) and self.__left.type != self.__right.type:
            raise TypeError(f"{self.__left.type} {self.__op} {self.__right.type} on line {self.lineno}")

        n1 = self.__left.interpret()
        n2 = self.__right.interpret()

        if isinstance(n1, Variable):
            n1 = n1.value

        if isinstance(n2, Variable):
            n2 = n2.value

        if   self.__op == '>':
            return n1 > n2

        elif self.__op == '>=':
            return n1 >= n2

        elif self.__op == '<':
            return n1 < n2

        elif self.__op == '<=':
            return n1 <= n2

        elif self.__op == '==':
            return n1 == n2

        elif self.__op == '!=':
            return n1 != n2

        elif self.__op == '+':
            if isinstance(n1, datetime):
                return n1 + timedelta(days = n2)
            else:
                return n1 + n2
        elif self.__op == '-':
            if isinstance(n1, datetime):
                return n1 - timedelta(days = n2)
            else:
                return n1 - n2

        elif self.__op == '*':
            if isinstance(n1, datetime):
                raise TypeError(f"Unsupported operation {self.__op} on {n1} on line {self.lineno}")
            return n1 * n2

        elif self.__op == '/':
            if isinstance(n1, datetime):
                raise TypeError(f"Unsupported operation {self.__op} on {n1} on line {self.lineno}")
            return n1 / n2

        elif self.__op == 'and':
            return n1 and n2

        elif self.__op == 'or':
            return n1 or n2

        else:
            raise SyntaxError(f"Invalid op {self.__op}")

    def __str__(self):
        return f"BinOp: '{self.__left}' '{self.__op}' '{self.__right}'"

    def __repr__(self):
        return self.__str__()
